"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
RFF 2：Extension Header Protocol(EHP)
Authors: jdh99 <jdh821@163.com>
"""

import utzpy.nlv1 as nlv1
import utzpy.common as common

import unittest

# 版本
EHP_VERSION_NAME = '1.1'

# 扩展头部
# 逐跳可选项头部
HEADER_HOP_BY_HOP_OPTIONS = 0x0
# 路由头部
HEADER_ROUTE = 0x1
# 分片头部
HEADER_FRAGMENT = 0x2
# 加密头部
HEADER_ENCRYPTION = 0x3
# FLP
HEADER_FLP = 0x4
# CMP
HEADER_CMP = 0x5
# WTS
HEADER_WTS = 0x5
# DUP
HEADER_DUP = 0x7
# SFTPA
HEADER_SFTP_A = 0x8
# SFTPB
HEADER_SFTP_B = 0x9
# STCP
HEADER_STCP = 0xA
# ITCP
HEADER_ITCP = 0xB
# DAP
HEADER_DAP = 0xC
# A类RSSI头部
HEADER_RSSI_A = 0xD
# B类RSSI头部
HEADER_RSSI_B = 0xE
# A类发送控制头部
HEADER_RADIO_TX_CONTROL_A = 0xF
# 125K激发RSSI头部
HEADER_EXCITER_RSSI = 0x10
# A类TOF定位头部
HEADER_TOF_LOCATION_A = 0x11
# A类DW1000接收头部
HEADER_DW1000_RX_A = 0x12
# A类DW1000发送完成头部
HEADER_DW1000_TX_END_A = 0x13
# A类坐标头部
HEADER_COORDINATE_A = 0x14
# 时间戳头部
HEADER_TIMESTAMP = 0x15
# 代理头部
HEADER_AGENT = 0x16
# 自组网控制头部
HEADER_ADHOCC = 0x17
# 压缩头部复合帧
HEADER_COMPRESS_COMPLEX = 0x18
# 标准头部复合帧
HEADER_STANDARD_COMPLEX = 0x19
# 全球单播地址分配服务器访问协议
HEADER_GUAAP = 0x1A
# 固定地址解析服务器访问协议
# 订阅服务器访问协议
# 简单安全头部
HEADER_SIMPLE_SECURITY = 0x1D
# 定长传输控制头部
HEADER_FIXED_LENGTH_TRANSFER_CONTROL = 0x1E
# 中继头部
HEADER_REPEAT = 0x1F
# 子分片头部
HEADER_SUB_FRAGMENT = 0x20
# 系统日志协议
HEADER_SLP = 0x22
# 设备间通信协议
HEADER_DCOM = 0x23
# 物联网终端协议
HEADER_ISH = 0x24


class RouteHeader:
    """路由头部结构"""

    def __init__(self):
        self.next_head = 0
        # 剩余路由数
        self.route_num = 0
        # 是否是严格源路由
        self.is_strict = False
        # 路由地址列表
        self.ia_list = list()


class SimpleSecurityHeader:
    """简单安全头部结构"""

    def __init__(self, next_head=0, pwd=''):
        self.next_head = next_head
        self.pwd = pwd


def bytes_to_route_header(data: bytearray) -> (RouteHeader, int):
    """
    字节流转换为路由头部
    :param data: 字节流data必须大于头部长度
    :return: 返回头部以及头部字节数.头部为None或者字节数为0表示转换失败
    """
    # 头部数据必须完整
    size = len(data)
    if size < 3:
        return None, 0

    header = RouteHeader()
    header.next_head = data[0]
    header_payload_len = data[1]
    if size < 2 + header_payload_len:
        return None, 0

    # 总长度不应该小于路由地址总长度
    header.route_num = data[2] & 0x7f
    header.is_strict = (data[2] >> 7) == 0x1
    if size < 3 + header.route_num * nlv1.IA_LEN:
        return None, 0

    header.ia_list.clear()
    offset = 3
    for i in range(header.route_num):
        header.ia_list.append(common.bytes_to_ia(data[offset:]))
        offset += nlv1.IA_LEN
    return header, offset


def route_header_to_bytes(header: RouteHeader) -> bytearray:
    """路由头部转换为字节流"""
    data = bytearray()
    data.append(header.next_head)
    # 头部数据长度
    header_payload_len = len(header.ia_list) * nlv1.IA_LEN + 1
    data.append(header_payload_len)

    route_num = header.route_num & 0x7f
    if header.is_strict:
        route_num |= 0x80
    data.append(route_num)

    for i in range(header.route_num):
        data += common.ia_to_bytes(header.ia_list[i])
    return data


def is_payload_header(head: int) -> bool:
    """是否载荷头部"""
    return (head == HEADER_FLP or head == HEADER_WTS or head == HEADER_DUP or head == HEADER_SFTP_A or
            head == HEADER_SFTP_B or head == HEADER_STCP or head == HEADER_ITCP or head == HEADER_DAP or
            head == HEADER_COMPRESS_COMPLEX or head == HEADER_STANDARD_COMPLEX or head == HEADER_GUAAP or
            head == HEADER_SLP or head == HEADER_DCOM or head == HEADER_ISH)


def bytes_to_simple_security_header(data: bytearray) -> (SimpleSecurityHeader, int):
    """
    字节流转换为简单安全头部
    :param data: 字节流data必须大于头部长度
    :return: 返回头部以及头部字节数.头部为None或者字节数为0表示转换失败
    """
    # 头部数据必须完整
    size = len(data)
    if size < 2:
        return None, 0

    header = SimpleSecurityHeader()
    header.next_head = data[0]
    header_payload_len = data[1]
    if size < 2 + header_payload_len:
        return None, 0

    header.pwd = data[2:2 + header_payload_len].decode('utf-8')
    return header, 2 + header_payload_len


def simple_security_header_to_bytes(header: SimpleSecurityHeader) -> bytearray:
    """简单安全头部转换为字节流"""
    data = bytearray()
    data.append(header.next_head)
    # 头部数据长度
    header_payload_len = len(header.pwd)
    data.append(header_payload_len)

    data += bytearray(header.pwd.encode())
    return data


class _UnitTest(unittest.TestCase):
    def test_router_header(self):
        header = RouteHeader()
        header.next_head = 2
        header.route_num = 3
        header.is_strict = True
        header.ia_list.append(0x1234567812345677)
        header.ia_list.append(0x1234567812345678)
        header.ia_list.append(0x1234567812345679)

        data = route_header_to_bytes(header)
        self.print_hex(data)
        self.assertEqual(data, bytearray(
            [0x02, 0x19, 0x83, 0x12, 0x34, 0x56, 0x78, 0x12, 0x34, 0x56, 0x77, 0x12, 0x34, 0x56, 0x78, 0x12, 0x34, 0x56,
             0x78, 0x12, 0x34, 0x56, 0x78, 0x12, 0x34, 0x56, 0x79]))

        header2, num = bytes_to_route_header(data)
        self.assertEqual(header2.next_head, 2)
        self.assertEqual(header2.route_num, 3)
        self.assertEqual(header2.is_strict, True)
        self.assertEqual(header2.ia_list[0], 0x1234567812345677)
        self.assertEqual(header2.ia_list[1], 0x1234567812345678)
        self.assertEqual(header2.ia_list[2], 0x1234567812345679)

    def test_simple_security_header(self):
        header = SimpleSecurityHeader(5, 'jdh99')
        data = simple_security_header_to_bytes(header)
        self.print_hex(data)
        self.assertEqual(data, bytearray([0x05, 0x05, 0x6a, 0x64, 0x68, 0x39, 0x39]))

        header2, num = bytes_to_simple_security_header(data)
        self.assertEqual(num, 7)
        self.assertEqual(header2.next_head, 5)
        self.assertEqual(header2.pwd, 'jdh99')

    @staticmethod
    def print_hex(data: bytearray):
        for i in data:
            print('0x%02x, ' % i, end='')
        print()


if __name__ == '__main__':
    suite = unittest.TestSuite()
    runner = unittest.TextTestRunner()
    runner.run(suite)



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
RFF 3：Fixed Length Transfer Protocol(FLP)
Authors: jdh99 <jdh821@163.com>
"""

import unittest

from crcmodbus import crc16

# 版本
FLP_VERSION_NAME = '1.2'

# 最大尾缀长度
FLP_SUFFIX_LEN_MAX = 0x7FF


def bytes_to_flp_frame(data: bytearray, is_need_crc: bool, fixed_len: int) -> bytearray:
    """字节流转换为FLP帧.fixedLen为0表示不是固定长度帧,不需要尾缀"""
    frame_len = 2 + len(data)
    if is_need_crc:
        frame_len += 2

    suffix_len = 0
    if frame_len < fixed_len:
        suffix_len = fixed_len - frame_len
        if suffix_len > FLP_SUFFIX_LEN_MAX:
            suffix_len = FLP_SUFFIX_LEN_MAX

    frame = bytearray()
    # 尾缀长度
    suffix_len_value = suffix_len
    if is_need_crc:
        suffix_len_value |= 0x8000

    frame.append((suffix_len_value >> 8) & 0xff)
    frame.append(suffix_len_value & 0xff)

    # crc校验
    if is_need_crc:
        frame.append(0)
        frame.append(0)

    # 正文
    frame += data
    # 尾缀
    frame += bytearray(suffix_len)

    if is_need_crc:
        crc = crc16.checksum(data)
        frame[2] = (crc >> 8) & 0xff
        frame[3] = crc & 0xff
    return frame


def flp_frame_to_bytes(frame: bytearray) -> bytearray:
    """FLP帧转换为字节流.字节流是FLP帧的数据正文.转换失败返回None"""
    size = len(frame)
    if size < 2:
        return None

    suffix_len_value = (frame[0] << 8) + frame[1]
    suffix_len = suffix_len_value & 0x7FFF
    is_need_crc = (suffix_len_value >> 15) == 0x1

    # 判断帧最小长度是否正确
    calc_frame_min_size = suffix_len + 2
    if is_need_crc:
        calc_frame_min_size = suffix_len + 2
    if size < calc_frame_min_size:
        return None

    if is_need_crc:
        crc_get = (frame[2] << 8) + frame[3]
        crc_calc = crc16.checksum(frame[4:size - suffix_len])
        if crc_get != crc_calc:
            return None
        return frame[4:size - suffix_len]
    else:
        return frame[2:size - suffix_len]


class _UnitTest(unittest.TestCase):
    def test_flp_header(self):
        data = bytearray([1, 2, 3, 4, 5])
        frame1 = bytes_to_flp_frame(data, True, 15)
        frame2 = bytes_to_flp_frame(data, False, 15)
        self.print_hex(frame1)
        self.print_hex(frame2)
        self.assertEqual(frame1, bytearray(
            [0x80, 0x06, 0x2a, 0xbb, 0x01, 0x02, 0x03, 0x04, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]))
        self.assertEqual(frame2, bytearray(
            [0x00, 0x08, 0x01, 0x02, 0x03, 0x04, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]))

        data1 = flp_frame_to_bytes(frame1)
        data2 = flp_frame_to_bytes(frame2)
        self.print_hex(data1)
        self.print_hex(data2)
        self.assertEqual(data1, bytearray([0x01, 0x02, 0x03, 0x04, 0x05]))
        self.assertEqual(data2, bytearray([0x01, 0x02, 0x03, 0x04, 0x05]))

    @staticmethod
    def print_hex(data: bytearray):
        for i in data:
            print('0x%02x, ' % i, end='')
        print()


if __name__ == '__main__':
    suite = unittest.TestSuite()
    runner = unittest.TextTestRunner()
    runner.run(suite)



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
RFF 1：Network Layer Version 1(NLv1)
Authors: jdh99 <jdh821@163.com>
"""

import utzpy.common as common

import unittest

# 协议族版本号
PROTOCOL_VERSION = 1

# 协议版本
NLV1_VERSION_NAME = "1.1"

# 头部长度
NLV1_HEAD_LEN = 22

# 特殊地址
# 未指定地址
IA_INVALID = 0x0
# 回环地址
IA_LOOPBACK = 0x1
# 广播地址
IA_BROADCAST = 0xFFFFFFFFFFFFFFFF

# IA地址字节数
IA_LEN = 8


class StandardHeader:
    """标准头部"""

    def __init__(self):
        self.version = 0
        self.frame_index = 0
        self.payload_len = 0
        self.next_head = 0
        self.hops_limit = 0
        self.src_ia = 0
        self.dst_ia = 0


_generate_frame_index = 0


def bytes_to_standard_header(data: bytearray) -> (StandardHeader, int):
    """
    字节流转换为标准头部.字节流是大端
    :param data: 字节流必须大于头部长度
    :return: 返回头部以及头部字节数.头部为None或者字节数为0表示转换失败
    """
    if len(data) < NLV1_HEAD_LEN:
        return None, 0

    header = StandardHeader()
    j = 0
    header.version = data[j]
    j += 1
    header.frame_index = data[j]
    j += 1
    header.payload_len = (data[j] << 8) + data[j + 1]
    j += 2
    header.next_head = data[j]
    j += 1
    header.hops_limit = data[j]
    j += 1
    header.src_ia = common.bytes_to_ia(data[j:])
    j += IA_LEN
    header.dst_ia = common.bytes_to_ia(data[j:])
    j += IA_LEN
    return header, j


def standard_header_to_bytes(header: StandardHeader) -> bytearray:
    """标准头部转换为字节流.字节流是大端"""
    data = bytearray()
    data.append(header.version)
    data.append(header.frame_index)
    data.append((header.payload_len >> 8) & 0xff)
    data.append(header.payload_len & 0xff)
    data.append(header.next_head)
    data.append(header.hops_limit)
    data += common.ia_to_bytes(header.src_ia)
    data += common.ia_to_bytes(header.dst_ia)
    return data


def is_global_ia(ia: int) -> bool:
    """是否是全球单播地址"""
    return ((ia >> 61) & 0xff) == 0x1


def is_constant_ia(ia: int) -> bool:
    """是否是固定单播地址"""
    return ((ia >> 61) & 0xff) == 0x2


def is_unique_local_ia(ia: int) -> bool:
    """是否是唯一本地地址"""
    return ((ia >> 57) & 0xff) == 0xfe


def generate_frame_index() -> int:
    """生成帧序号"""
    global _generate_frame_index

    _generate_frame_index += 1
    if _generate_frame_index > 0xff:
        _generate_frame_index = 0
    return _generate_frame_index


class _UnitTest(unittest.TestCase):
    def test_bytes_to_standard_header(self):
        data = bytearray(
            [0x01, 0x05, 0x00, 0x06, 0x07, 0x08, 0x12, 0x34, 0x56, 0x78, 0x12, 0x34, 0x56, 0x78, 0x87, 0x65, 0x43, 0x21,
             0x87, 0x65, 0x43, 0x21])
        header, offset = bytes_to_standard_header(data)
        self.assertEqual(offset, NLV1_HEAD_LEN)
        self.assertEqual(header.version, PROTOCOL_VERSION)
        self.assertEqual(header.frame_index, 5)
        self.assertEqual(header.payload_len, 6)
        self.assertEqual(header.next_head, 7)
        self.assertEqual(header.hops_limit, 8)
        self.assertEqual(header.src_ia, 0x1234567812345678)
        self.assertEqual(header.dst_ia, 0x8765432187654321)

    def test_standard_header_to_bytes(self):
        header = StandardHeader()
        header.version = PROTOCOL_VERSION
        header.frame_index = 5
        header.payload_len = 6
        header.next_head = 7
        header.hops_limit = 8
        header.src_ia = 0x1234567812345678
        header.dst_ia = 0x8765432187654321
        data = standard_header_to_bytes(header)
        self.assertEqual(data, bytearray(
            [0x01, 0x05, 0x00, 0x06, 0x07, 0x08, 0x12, 0x34, 0x56, 0x78, 0x12, 0x34, 0x56, 0x78, 0x87, 0x65, 0x43, 0x21,
             0x87, 0x65, 0x43, 0x21]))

    def test_ia(self):
        self.assertEqual(is_global_ia(0x2140000000000401), True)
        self.assertEqual(is_global_ia(0x4001000000000401), False)
        self.assertEqual(is_constant_ia(0x2140000000000401), False)
        self.assertEqual(is_constant_ia(0x4001000000000401), True)

    @staticmethod
    def print_hex(data: bytearray):
        for i in data:
            print('%02x ' % i, end='')
        print()


if __name__ == '__main__':
    suite = unittest.TestSuite()
    runner = unittest.TextTestRunner()
    runner.run(suite)



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
通用模块
Authors: jdh99 <jdh821@163.com>
"""

import utzpy.nlv1 as nlv1

import unittest


def bytes_to_ia(data: bytearray) -> int:
    """从字节流中取出IA地址.字节流是大端"""
    if len(data) < nlv1.IA_LEN:
        return 0

    ia = 0
    for i in range(nlv1.IA_LEN):
        ia += data[i] << ((nlv1.IA_LEN - 1 - i) << 3)
    return ia


def ia_to_bytes(ia: int) -> bytearray:
    """将IA地址转换为字节流.字节流是大端"""
    data = bytearray()
    for i in range(nlv1.IA_LEN):
        data.append((ia >> ((nlv1.IA_LEN - 1 - i) << 3)) & 0xff)
    return data


class _UnitTest(unittest.TestCase):
    def test_bytes_to_ia(self):
        ia = bytes_to_ia(bytearray([0x21, 0x40, 0x00, 0x00, 0x00, 0x00, 0x04, 0x01]))
        self.assertEqual(ia, 0x2140000000000401)

    def test_ia_to_bytes(self):
        data = ia_to_bytes(0x2140000000000401)
        self.assertEqual(data, bytearray([0x21, 0x40, 0x00, 0x00, 0x00, 0x00, 0x04, 0x01]))


if __name__ == '__main__':
    suite = unittest.TestSuite()
    runner = unittest.TextTestRunner()
    runner.run(suite)



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
RFF 4：Control Message Protocol(CMP)
Authors: jdh99 <jdh821@163.com>
"""

# 版本
CMP_VERSION_NAME = '1.1'

# 消息类型
# 申请从机
CMP_MSG_TYPE_REQUEST_SLAVE_ROUTER = 0x88
# 分配从机
CMP_MSG_TYPE_ASSIGN_SLAVE_ROUTER = 0x89
# 连接父路由
CMP_MSG_TYPE_CONNECT_PARENT_ROUTER = 0x8A
# 确认连接父路由
CMP_MSG_TYPE_ACK_CONNECT_PARENT_ROUTER = 0x8B
